print("import: 'quantecon'")
import quantecon

print("import: 'quantecon.tests'")
import quantecon.tests

print("import: 'quantecon.game_theory'")
import quantecon.game_theory

print("import: 'quantecon.markov'")
import quantecon.markov

print("import: 'quantecon.random'")
import quantecon.random

print("import: 'quantecon.util'")
import quantecon.util

