"""
Warnings Module
===============

Contains a collection of warning messages for consistent package wide notifications

"""

#-Numba-#
numba_import_fail_message = ("Numba import failed. Falling back to non-optimized routines.\n" 
                            "This will reduce the overall performance of this package.\n"    
                            "To install please use the anaconda distribution.\n"             
                            "http://continuum.io/downloads")
